/*
* Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
*
* SPDX-License-Identifier:  GPL-2.0+
*/

#include <common.h>
#include <ring_platform_features.h>
#include "asm/arch/mach/platform.h"
#include <../drivers/mstar/aesdma/drvAESDMA.h>
#include <bootm.h>
#include <malloc.h>
#include <nand.h>
#include <fb_otp.h>
#include <spinand.h>
#include <drvSPINAND.h>
#include <unlocker.h>
#include <cli.h>
#include <fastboot.h>
#include <ubi_uboot.h>

#ifdef DEBUG
#define debugf(fmt, args...) do { printf("%s(): ", __func__); printf(fmt, ##args); } while (0)
#else
#define debugf(fmt, args...)
#endif

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

#ifndef COCOA_KERNEL_SIZE
#error Need to define kernel image size in COCOA_KERNEL_SIZE
#endif

#ifndef COCOA_RTOS_SIZE
#error Need to define rtos image size in COCOA_RTOS_SIZE
#endif

#ifndef COCOA_UBI_MTD
#error Need to define ubi mtd number in COCOA_UBI_MTD
#endif

#ifndef COCOA_RTOS_PART_NAME
#error Need to define rtos image partition name
#endif

#ifndef COCOA_RTOS_BACKUP_PART_NAME
#error Need to define rtos backup image partition name
#endif

#ifndef COCOA_RECOVERY_KERNEL_PART_NAME
#error Need to define recovery kernel image partition name
#endif

#ifndef COCOA_RECOVERY_BACKUP_KERNEL_PART_NAME
#error Need to define recovery backup kernel image partition name
#endif

#ifndef COCOA_RTOS_BOOT_ADDRESS
#error Need to define the RTOS boot address
#endif

#ifndef COCOA_RECOVERY_KERNEL_LOAD_ADDRESS
#error Need to define the recovery kernel load address
#endif

#ifndef COCOA_USERDATA_SIZE
#error Need to specify userdata size in bytes
#endif

#ifndef COCOA_DATA_PARTITION
#error Need to specify data partition name
#endif

#ifndef COCOA_RECOVERY_SIZE
#error Need to specify recovery size in bytes
#endif

#define COCOA_UBI_NAME "UBI"
#define COCOA_RECOVERY_PARTITION "recovery"

#define NOCONSOLE_ARG     "console= "
#define NOCONSOLE_ARG_LEN (sizeof(NOCONSOLE_ARG) - 1)

#define UNLOCKED_KERNEL_ROOT "ringboot.unlocked_kernel="
#define SERIALNO_ROOT "ringboot.serialno="
#define BOARDID_ROOT " ringboot.boardid="

#define LOCKED_KERNEL_ARG       UNLOCKED_KERNEL_ROOT "false "
#define LOCKED_KERNEL_ARG_LEN   (sizeof(LOCKED_KERNEL_ARG) - 1)
#define UNLOCKED_KERNEL_ARG     UNLOCKED_KERNEL_ROOT "true "
#define UNLOCKED_KERNEL_ARG_LEN (sizeof(UNLOCKED_KERNEL_ARG) - 1)

#define BACKUP_KERNEL_STR "ringboot.backup_kernel"
#define BACKUP_KERNEL_STR_LEN (sizeof(BACKUP_KERNEL_STR) - 1)

// Maximum number of repeated recovery boots allowed
#define MAX_RECOVERY_MODE_BOOT_COUNT 5

/* This is intended as a failsafe against programming errors which inadvertently
 * skip authentication and hence allow boot.  Unless runAuthenticate has returned
 * successfully, this value should not be set in the "has_authenticated"
 * variable
 */
#define HAS_AUTHENTICATED_KEY   0x41757468
#define NO_AUTH_KEY             0x44657641
static uint32_t has_authenticated = 0;

#define PRIV_BOOTARGS_LEN 1024
static char *private_bootargs = NULL;

static char dsn[COCOA_DSN_LEN+1];
static char boardid[COCOA_BOARD_ID_LEN+1];
static bool read_otp = false;

static const unsigned int BOARD_ID_PROD_ID_INDEX; // = 0
static const unsigned int BOARD_ID_PROD_ID_LEN      =  4;
static const unsigned int BOARD_ID_2ND_SOURCE_INDEX =  4;
static const unsigned int BOARD_ID_MAJOR_REV_INDEX  =  7;
static const unsigned int BOARD_ID_MINOR_REV_INDEX  =  8;

static const char * const BOARD_ID_PROD_ID_STRS[] = {
	"c0ca",
	"c0cb",
	"c0cc",
	"c0cd"
};

enum RB_PROD_ID_TYPES {
	RB_PRODUCT_COCOA,
	RB_PRODUCT_SMORE,
	RB_PRODUCT_GOLDFISH,
	RB_PRODUCT_RINGTERCOM,
	RB_PRODUCT_UNKNOWN
};

static const char *ps5250_str = "ps5250";
static const char *sp2309_str = "sp2309";

enum RB_MAJOR_REVS {
	RB_MAJOR_REV_UNKNOWN = 0,

	RB_MAJOR_REV_MIN = '0',

	RB_MAJOR_REV_PROTO = RB_MAJOR_REV_MIN,
	RB_MAJOR_REV_HVT,
	RB_MAJOR_REV_EVT,
	RB_MAJOR_REV_DVT,
	RB_MAJOR_REV_PVT,
	RB_MAJOR_REV_MP,

	RB_MAJOR_REV_MAX
};


typedef struct rtk_image_header {
	u32     ivt[8];     // Interrupt Vector Table, 32byte
	u32     marker;     // RTK header, 0x5F4B5452 (“RTK_”)
	u32     size;       // Image size (include header)
	u32     checksum;   // Checksum (exclude header)
	u32     reserved;   // Reserved
} rtk_image_header_t;

void signal_boot_ack_pulse(void);
void configure_boot_ack(void);

static void alloc_bootargs(void)
{
	if (!private_bootargs) {
		private_bootargs = malloc(PRIV_BOOTARGS_LEN);
		if (private_bootargs)
			private_bootargs[0] = '\0';
	}
}

char *get_bootargs(void)
{
	alloc_bootargs();
	return private_bootargs;
}

int set_bootargs(const char *value)
{
	alloc_bootargs();
	if (private_bootargs == NULL)
		return -1;
	strlcpy(private_bootargs, value, PRIV_BOOTARGS_LEN);
	return 0;
}

#ifdef MTDPARTS_ON_KERNEL_CMDLINE
// returns the safe number of characters that should be used from mtdparts
// protects against something invalid being added to the kernel command line
static int mtd_parts_length(const char *mtdparts)
{
	const char key[] = "mtdparts=";
	int sz;
	if (strncmp(mtdparts, key, sizeof(key) - 1) != 0) {
		return 0;
	}
	for(sz = sizeof(key) - 1; sz < MTDPARTS_MAXLEN; sz++) {
		if (mtdparts[sz] == 0 || mtdparts[sz] == ' ') {
			return sz;
		}
	}
	return MTDPARTS_MAXLEN;
}
#endif

static void set_ring_recovery_command_line(unlock_state_t unlock_state, bool use_backup)
{
	const char args[] = "ipc.disable_rtos=1 amazon_sdio.disable_comms=1 rdinit=/init_recovery ringboot.validated_kernel=true ";
	int err = 0;
#ifdef MTDPARTS_ON_KERNEL_CMDLINE
	char *mtdparts = getenv("mtdparts");
#endif

	strcpy(get_bootargs(), args);
	if (unlock_state < UNLOCK_STATE_RESTRICTED_UNLOCK) {
		strncat(get_bootargs(), NOCONSOLE_ARG, NOCONSOLE_ARG_LEN);
		strncat(get_bootargs(), LOCKED_KERNEL_ARG, LOCKED_KERNEL_ARG_LEN);
	} else {
		strncat(get_bootargs(), UNLOCKED_KERNEL_ARG, UNLOCKED_KERNEL_ARG_LEN);
	}
	if (!read_otp) {
		err = try_otp_read_all(dsn, boardid);
	}
	if (err == 0) {
		read_otp = true;
		strncat(get_bootargs(), SERIALNO_ROOT, sizeof(SERIALNO_ROOT)-1);
		strncat(get_bootargs(), dsn, COCOA_DSN_LEN);

		strncat(get_bootargs(), BOARDID_ROOT, sizeof(BOARDID_ROOT)-1);
		strncat(get_bootargs(), boardid, COCOA_BOARD_ID_LEN);
	}

#ifdef MTDPARTS_ON_KERNEL_CMDLINE
	strncat(get_bootargs(), " ", 1);
	strncat(get_bootargs(), mtdparts, mtd_parts_length(mtdparts));
#endif

	if (use_backup) {
		strncat(get_bootargs(), " ", 1);
		strncat(get_bootargs(), BACKUP_KERNEL_STR, BACKUP_KERNEL_STR_LEN);
	}

	printf("Recovery command line: %s\n", get_bootargs());
}

int read_otp_page(int page, u_char *addr, size_t size_to_read, int dev)
{
	nand_info_t *nand = &nand_info[dev];
	loff_t off = 0;
	int err = -1;

	size_t pagesize = nand->writesize;
	off += (page + MDrv_SPINAND_GetOTPPageOffset())* pagesize;

	if (size_to_read <= pagesize) {
		MDrv_SPINAND_EnableOtp(true);
		err = nand_read_skip_bad(nand, off, &size_to_read, NULL,
			pagesize, addr);
		MDrv_SPINAND_EnableOtp(false);
	}

	return err;
}
extern int get_part(const char *partname, int *idx, loff_t *off, loff_t *size,
		loff_t *maxsize);
static int read_nand_partition(const char *part_name, size_t size_to_read,
								u_char *addr, int dev)
{
	nand_info_t *nand = &nand_info[dev];
	int idx;
	loff_t off;
	loff_t part_size;
	loff_t maxsize;
	int err = -1;

	get_part(part_name, &idx, &off, &part_size, &maxsize);

	if (size_to_read <= part_size) {
		err = nand_read_skip_bad(nand, off, &size_to_read, NULL,
			part_size, addr);
	}
	return err;
}

#define authfail(err) \
	do { \
		printf("Authentication failed! (err=%d)\n", err); \
		goto exit_authfail; \
	} while(0);

static int authenticate_image(u32 image_address, u32 image_size,
                               u32 key_offset, unlock_state_t unlock_state)
{
	int err;
	u8* ringkey = (u8*)KEY_CUST_LOAD_ADDRESS;

	err = load_ring_public_key(ringkey, PROD_KEY_PAGE, key_offset,
	                           RSA_KEYSIZE);
	if (err)
		authfail(err);

	if (runAuthenticate((U32)image_address, image_size, (U32*)ringkey)) {
		has_authenticated = HAS_AUTHENTICATED_KEY;
	} else {
		if (unlock_state < UNLOCK_STATE_RESTRICTED_UNLOCK) {
			err = -1;
			authfail(err);
		} else {
			printf("Attempt authentication with dev keys\n");
			err = load_ring_public_key(ringkey, DEV_KEY_PAGE,
			                           key_offset, RSA_KEYSIZE);
			if (err) {
				authfail(err);
			} else {
				if (runAuthenticate((U32)image_address,
				                     image_size,
				                     (U32*)ringkey)) {
					has_authenticated = HAS_AUTHENTICATED_KEY;
				} else if (unlock_state < UNLOCK_STATE_NON_PROD) {
					err = -2;
					authfail(err);
				} else {
					printf("Non-production device; allowing boot to continue!\n");
					has_authenticated = NO_AUTH_KEY;
				}
			}
		}
	}

	if (err == 0) {
		printf("Authenticate image passed!\n");
		return 0;
	}

exit_authfail:
	return err;
}

static bool can_boot(int err, unlock_state_t unlock_state)
{
	if (err) {
		goto no_boot;
	}
	if (has_authenticated == HAS_AUTHENTICATED_KEY) {
		goto ok_to_boot;
	} else if ((has_authenticated == NO_AUTH_KEY) &&
	           (unlock_state == UNLOCK_STATE_NON_PROD)) {
		goto ok_to_boot;
	}
no_boot:
	return false;
ok_to_boot:
	return true;
}

static int boot_recovery_kernel(unlock_state_t unlock_state, bool use_backup)
{
	int err = -1;
	u_char *kernel_load_address =
				(u_char *)COCOA_RECOVERY_KERNEL_LOAD_ADDRESS;
	static char *argv[1];
	u32 kernel_size;

	set_ring_recovery_command_line(unlock_state, use_backup);
	err = read_nand_partition(use_backup ? STR(COCOA_RECOVERY_BACKUP_KERNEL_PART_NAME) : STR(COCOA_RECOVERY_KERNEL_PART_NAME),
	                          COCOA_KERNEL_SIZE,
	                          kernel_load_address, 0);

	if (err) {
		printf("Failed to read kernel image from NAND\n");
		goto exit;
	}

	kernel_size = image_get_image_size((image_header_t *)kernel_load_address);
	if (kernel_size > COCOA_KERNEL_SIZE) {
		debugf("Invalid kernel image size\n");
		err = -2;
		goto exit;
	}

	err = authenticate_image((u32)kernel_load_address, kernel_size,
	                         KERN_KEY_OFFSET, unlock_state);

	if (can_boot(err, unlock_state)) {
		signal_boot_ack_pulse();
		argv[0] = STR(COCOA_RECOVERY_KERNEL_LOAD_ADDRESS);
		/* Perform bootm processing */
		err = do_bootm_states(NULL, 0, 1, argv,
				BOOTM_STATE_START | BOOTM_STATE_FINDOS |
				BOOTM_STATE_FINDOTHER | BOOTM_STATE_LOADOS |
				BOOTM_STATE_OS_PREP | BOOTM_STATE_OS_FAKE_GO |
				BOOTM_STATE_OS_GO,
				&images, 1);
	}

exit:
	printf("Unexpected exit from %s: err=%d has_authenticated=0x%x!\n", __func__, err, has_authenticated);
	return err;
}



static int boot_rtos(unlock_state_t unlock_state, bool use_backup)
{
	int err = -1;
	u_char *rtos_boot_address = (u_char *)COCOA_RTOS_BOOT_ADDRESS;
	u32 rtos_size;

	err = read_nand_partition(use_backup ? STR(COCOA_RTOS_BACKUP_PART_NAME) : STR(COCOA_RTOS_PART_NAME), COCOA_RTOS_SIZE,
	                          rtos_boot_address, 0);

	if (err) {
		printf("Failed to read RTOS image from NAND\n");
		goto exit;
	}

	rtos_size = ((rtk_image_header_t *)rtos_boot_address)->size;
	if (rtos_size > COCOA_RTOS_SIZE) {
		debugf("Invalid rtos image size\n");
		err = -2;
		goto exit;
	}

	err = authenticate_image((u32)rtos_boot_address, rtos_size,
	                         RTOS_KEY_OFFSET, unlock_state);
	if (can_boot(err, unlock_state)) {
		/* The nand driver appears not to invalidate cache ranges, so
		   push this cache enable to the last possible moment. */
		dcache_enable();
		icache_enable();
		signal_boot_ack_pulse();
		((void *(*)(void))rtos_boot_address)();
		dcache_disable();
		icache_disable();
		/* Continuing here is a (highly unlikely) error case. Allow it
		   to be detected */
		err = -6;
	}

exit:
	printf("Unexpected exit from %s: err=%d has_authenticated=0x%x!\n", __func__, err, has_authenticated);
	return err;
}

extern int unlocker_get_challenge(char *unlock_code, int unlock_code_len);
extern int unlocker_write_signature(uint8_t *unlock_signature, int unlock_sig_len);
extern int unlocker_relock_device(bool reboot);

int get_unlock_code(char *unlock_code, int unlock_code_len)
{
	return unlocker_get_challenge(unlock_code, unlock_code_len);
}

int get_lock_state(char *lock_state, int lock_state_len)
{
	int err = 0;
	unlock_state_t unlock_state;
	is_unlocked(&unlock_state);

	switch (unlock_state) {
		case UNLOCK_STATE_LOCKED:
			strlcpy(lock_state, "UNLOCK_STATE_LOCKED", lock_state_len);
			break;

		case UNLOCK_STATE_RESTRICTED_UNLOCK:
			strlcpy(lock_state, "UNLOCK_STATE_RESTRICTED_UNLOCK", lock_state_len);
			break;

		case UNLOCK_STATE_FULL_UNLOCK:
			strlcpy(lock_state, "UNLOCK_STATE_FULL_UNLOCK", lock_state_len);
			break;

		case UNLOCK_STATE_NON_PROD:
			strlcpy(lock_state, "UNLOCK_STATE_NON_PROD", lock_state_len);
			break;

		case UNLOCK_STATE_INVALID:
		default:
			/* Invalid state - return error */
			printf("Invalid unlock state %d\n", unlock_state);
			err = -1;
	}
	return err;
}

int write_unlock_signature(uint8_t *unlock_signature, int unlock_sig_len)
{
	return unlocker_write_signature(unlock_signature, unlock_sig_len);
}

int relock_device(void)
{
	return unlocker_relock_device(false);
}

/* This function is in cmd_nand.c  */
extern int read_nand_partition(const char *part_name, size_t size_to_read,
                               u_char *addr, int dev);


int load_ring_public_key(u8* key_load_addr, int key_page, int key_offset, int key_size)
{
	int err = -1;
	static const int load_size = 2048; /* 1 NAND page */
	u8 *tmpkey = malloc(load_size);
	if (!tmpkey) {
		printf("Failed to allocate space to load key\n");
		goto out;
	}
	if (key_offset + key_size > load_size) {
		printf("Requesting key outside page area!\n");
		goto out_free;
	}

	err = read_otp_page(key_page, tmpkey, load_size, 0);
	if (err) {
		printf("Failed to read ring public key from NAND OTP\n");
		goto out_free;
	}
	memcpy(key_load_addr, tmpkey + key_offset, key_size);
out_free:
	free(tmpkey);
out:
	return err;
}

static long getenv_long(const char *name, int base, long default_val)
{
	const char *str = getenv(name);

	return str ? simple_strtol(str, NULL, base) : default_val;
}

char *getserial(void)
{
	int err = 0;
	char *dsn_p;

	if (!read_otp)
		err = try_otp_read_all(dsn, boardid);

	if (err == 0) {
		read_otp = true;
		dsn_p = dsn;
	} else {
		dsn_p = getenv("serial#");
	}
	return dsn_p;
}

const char *getsensor(void)
{
	int err = 0;
	const char *sensor = ps5250_str;

	if (!read_otp)
		err = try_otp_read_all(dsn, boardid);

	if (err == 0) {
		enum RB_PROD_ID_TYPES product_type = RB_PRODUCT_UNKNOWN;
		enum RB_MAJOR_REVS    major_rev = RB_MAJOR_REV_UNKNOWN;
		int second_source_ver;
		int minor_rev;

		read_otp = true;
		for (int i = 0; i < ARRAY_SIZE(BOARD_ID_PROD_ID_STRS); i++) {
			if (strncmp(&boardid[BOARD_ID_PROD_ID_INDEX],
				    BOARD_ID_PROD_ID_STRS[i],
				    BOARD_ID_PROD_ID_LEN) == 0) {
				product_type = i;
				break;
			}
		}
		if (boardid[BOARD_ID_MAJOR_REV_INDEX] >= RB_MAJOR_REV_MIN &&
		    boardid[BOARD_ID_MAJOR_REV_INDEX] < RB_MAJOR_REV_MAX)
			major_rev = boardid[BOARD_ID_MAJOR_REV_INDEX];

		second_source_ver =
			(boardid[BOARD_ID_2ND_SOURCE_INDEX] - '0') * 10 +
			(boardid[BOARD_ID_2ND_SOURCE_INDEX + 1] - '0');

		minor_rev = boardid[BOARD_ID_MINOR_REV_INDEX] - '0';
		switch (product_type) {
		case RB_PRODUCT_SMORE:
			if (second_source_ver > 0)
				sensor = sp2309_str;
			break;
		case RB_PRODUCT_GOLDFISH:
			/* Goldfish is switching to SP2309 from DVT2.1 (DVT3 in internal numbering) */
			if ((major_rev > RB_MAJOR_REV_DVT) ||
			    ((major_rev == RB_MAJOR_REV_DVT) && (minor_rev >= 3)))
				sensor = sp2309_str;
			break;
		default:
			break;
		}
	}
	return sensor;
}

/* Forward declarations for functions to recreate customer partition */
int ubifs_init(void);
int uboot_ubifs_mount(char *vol_name);
int ubi_remove_vol(char *volume);
int ubi_create_vol(char *volume, int64_t size, int dynamic);

/* Check the recovery rootfs
 * Returns true if the recovery rootfs appears to be bootable
 */
static bool check_recovery(void)
{
	// Can recovery be mounted?
	if (uboot_ubifs_mount("ubi:" COCOA_RECOVERY_PARTITION) != 0) {
		printf("Recovery rootfs is corrupt\n");

		// Try creating the partition in case it's missing
		// the update process should rewrite the whole volume
		ubi_create_vol(COCOA_RECOVERY_PARTITION, COCOA_RECOVERY_SIZE, 1);
		return false;
	}
	return true;
}

/* Erase an mtd partition
 * Returns true if this succeeds
 */
static bool erase_nand(const char *part_name, int dev)
{
	int idx;
	loff_t max_size;
	nand_erase_options_t opts;

	memset(&opts, 0, sizeof(opts));
	if (get_part(part_name, &idx, &opts.offset, &opts.length, &max_size) != 0) {
		printf("erase_nand failed to find part\n");
		return false;
	}

	printf("erase_nand device %d offset 0x%llx, size 0x%llx\n", dev, opts.offset, opts.length);
	nand_info_t *nand = &nand_info[dev];
	int ret = nand_erase_opts(nand, &opts);
	if (ret != 0) {
		printf("erase_nand failed %d\n", ret);
		return false;
	}
	return true;
}

/* Normal boot;
 * nand read.e COCOA_RTOS_BOOT_ADDRESS COCOA_RTOS_PART_NAME COCOA_RTOS_SIZE
 * go COCOA_RTOS_BOOT_ADDRESS
 *
 * nand read.e COCOA_RECOVERY_KERNEL_LOAD_ADDRESS COCOA_RECOVERY_KERNEL_PART_NAME COCOA_KERNEL_SIZE
 * bootm COCOA_RECOVERY_KERNEL_LOAD_ADDRESS
 */
int do_ring_autoboot(void)
{
	unlock_state_t unlock_state;
	ring_config config;
	configure_boot_ack();
	signal_boot_ack_pulse();
#ifdef CONFIG_AUTO_TRY_FASTBOOT
#ifndef CONFIG_AUTO_TRY_FASTBOOT_TIMEOUT_MS
#define CONFIG_AUTO_TRY_FASTBOOT_TIMEOUT_MS 2000
#endif

	/* Negative value will skip fastboot. 0 will remain in fastboot */
	static long fastbootwaitms = -1;

	if (fastbootwaitms < 0)
		fastbootwaitms = getenv_long("fastbootwaitms", 10,
		                             CONFIG_AUTO_TRY_FASTBOOT_TIMEOUT_MS);

	if (fastbootwaitms >= 0)
		fastboot_try(fastbootwaitms);
#endif

	is_unlocked(&unlock_state);

	/* Locked devices don't get access to the command line */
	if ((unlock_state > UNLOCK_STATE_LOCKED) && had_ctrlc()) {
		cli_init();
		cli_loop();
	}

	bool erased_ubi = false;

	// Load ubi partition
	if (ubi_part(COCOA_UBI_NAME, NULL) != 0) {
		printf("failed to read UBI\n");

		// This should recreate ubi in a good state
		erase_nand(COCOA_UBI_NAME, 0);
		ubi_part(COCOA_UBI_NAME, NULL);
		erased_ubi = true;
	}

	// This will recreate bootconfig if needed
	get_ring_config(&config);

	if (unlock_state == UNLOCK_STATE_RESTRICTED_UNLOCK) {
		platform_increment_boot_count();
		/* Re-evaluate unlock state */
		is_unlocked(&unlock_state);
	}

	if (erased_ubi) {
		// Recreate the recovery partition
		check_recovery();
	}

	/* Recover/recreate customer partition */
	unsigned long before_ubi = get_timer(0);
	ubifs_init();
	int userdata_corrupt = (uboot_ubifs_mount("ubi:" STR(COCOA_DATA_PARTITION)) != 0);
	int factory_reset = (config.boot.rebootReason == REBOOT_REASON_FACTORY_RESET);
	if (userdata_corrupt || factory_reset) {
		printf("Recreating customer partition corrupt=%d factory_reset=%d\n", userdata_corrupt, factory_reset);
		ubi_remove_vol(STR(COCOA_DATA_PARTITION));
		ubi_create_vol(STR(COCOA_DATA_PARTITION), COCOA_USERDATA_SIZE, 1);
	}
	printf("%lums to manage ubifs\n", get_timer(before_ubi));

	if (factory_reset) {
		/* Clear any unlock codes before we try to boot further */
		relock_device();
		/* Re-evaluate unlock state */
		is_unlocked(&unlock_state);

		/* Reset reboot reason */
		config.boot.recoveryMode = 0;
		config.boot.rebootReason = REBOOT_REASON_NORMAL;
		set_ring_config(&config);
	}

	if (config.boot.recoveryMode) {

		// Keep a count of recovery boots
		config.boot.recoveryBootCount++;
		printf("Recovery Boot count %d/%d\n", config.boot.recoveryBootCount, MAX_RECOVERY_MODE_BOOT_COUNT);

		if (config.boot.recoveryBootCount <= MAX_RECOVERY_MODE_BOOT_COUNT && check_recovery()) {
			set_ring_config(&config);
			boot_recovery_kernel(unlock_state, false); // no return on success
			printf("Trying backup recovery kernel\n");
			boot_recovery_kernel(unlock_state, true); // try the backup kernel
		}

		/* If the recovery mode kernel is unbootable,
		 * or we've tried to boot too many times into recovery
		 * clear state which forces boot into that mode */
		config.boot.recoveryMode = 0;
		config.boot.rebootReason = (config.boot.rebootReason == REBOOT_REASON_APPLY_OTA) ? REBOOT_REASON_FAILED_OTA : REBOOT_REASON_NORMAL;
		set_ring_config(&config);
	}

	boot_rtos(unlock_state, false); // no return on success
	printf("Trying backup rtos\n");
	boot_rtos(unlock_state, true); // Try the backup rtos

	if (unlock_state < UNLOCK_STATE_NON_PROD) {
		/* Production device; halt on boot error */
		printf("HALT\n");
		for (;;) ;
	}

#ifdef CONFIG_AUTO_TRY_FASTBOOT
	/* We failed to boot.  On retry just stay in fastboot */
	fastbootwaitms = 0;
	clear_ctrlc();
#endif

	return 0;
}
