// SPDX-License-Identifier: BSD-2-Clause
/*
 * Copyright (C) 2016 The Android Open Source Project
 */

#include <common.h>
#include <fastboot.h>
#include <fastboot-internal.h>
#include <fb_mmc.h>
#include <fb_nand.h>
#include <fb_otp.h>
#include <part.h>
#include <stdlib.h>
#include <ubi_uboot.h>
#include <../mstar/gpio/drvGPIO.h>
#include <unlocker.h>

/**
 * image_size - final fastboot image size
 */
static u32 image_size;

/**
 * fastboot_bytes_received - number of bytes received in the current download
 */
static u32 fastboot_bytes_received;

/**
 * fastboot_bytes_expected - number of bytes expected in the current download
 */
static u32 fastboot_bytes_expected;

static void okay(char *, char *);
static void getvar(char *, char *);
static void download(char *, char *);
#ifdef CONFIG_FASTBOOT_FLASH
static void flash(char *, char *);
static void erase(char *, char *);
#endif
static void reboot_bootloader(char *, char *);
#ifdef CONFIG_FASTBOOT_CMD_OEM
static void oem_format(char *, char *);
static void oem_ubi(char *, char *);
static void oem_coda(char *, char *);
static void oem_setserial(char *, char *);
static void oem_setvar(char *, char *);
static void oem_relock(char *, char *);
static void oem_setfactory(char *, char *);
#ifdef CONFIG_FASTBOOT_FLASH_NAND
static void oem_otp_nand_lock(char *cmd_parameter, char *response);
#endif
static void handle_oem_command(char *, char *);
#endif

#ifdef CONFIG_FASTBOOT_FLASH_NAND
extern int writeSpinandCIS(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[]);
extern int write_unlock_signature(uint8_t *unlock_signature, int unlock_sig_len);
#endif


struct cmd_entry {
	const char *command;
	void (*dispatch)(char *cmd_parameter, char *response);
};

static int handle_commands_internal(const struct cmd_entry *cmds,
                                    int num_commands,
                                    char *cmd_string,
                                    char *response);

#define FB_WHITELIST_BASE	5	// getvar, boot, continue, reboot,
	                                // reboot-bootloader

#ifdef CONFIG_FASTBOOT_CMD_OEM
#define FB_WHITELIST_OEM	1
#else
#define FB_WHITELIST_OEM	0
#endif

#ifdef CONFIG_FASTBOOT_FLASH
#define FB_WHITELIST_FLASH	2	// flash, download (unlock handled in func flash requires both)
#else
#define FB_WHITELIST_FLASH	0
#endif

#define FB_WHITELIST (FB_WHITELIST_BASE + FB_WHITELIST_OEM + FB_WHITELIST_FLASH)
static const struct cmd_entry commands[] = {
	/* These are ordered according to the enum in fastboot.h.  They have
	   special handling elsewhere */
	[FASTBOOT_COMMAND_GETVAR] = {
		.command = "getvar",
		.dispatch = getvar
	},
	[FASTBOOT_COMMAND_BOOT] = {
		.command = "boot",
		.dispatch = okay
	},
	[FASTBOOT_COMMAND_CONTINUE] = {
		.command = "continue",
		.dispatch = okay
	},
	[FASTBOOT_COMMAND_REBOOT] = {
		.command = "reboot",
		.dispatch = okay
	},
	[FASTBOOT_COMMAND_REBOOT_BOOTLOADER] = {
		.command = "reboot-bootloader",
		.dispatch = reboot_bootloader
	},
	/* Below here are in array order and are handled in this file */
#ifdef CONFIG_FASTBOOT_CMD_OEM
	{
		.command = "oem",
		.dispatch = handle_oem_command,
	},
#endif
#ifdef CONFIG_FASTBOOT_FLASH
	{
		.command = "flash",
		.dispatch = flash
	},
	{
		.command = "download",
		.dispatch = download
	},
	{
		.command = "erase",
		.dispatch = erase
	},
#endif
	{
		.command = "set_active",
		.dispatch = okay
	},
};

#ifdef CONFIG_FASTBOOT_CMD_OEM
#define OEM_WHITELIST	1	// Only oem_relock is allowed.
static const struct cmd_entry oem_commands[] = {
	{
		.command = "relock",
		.dispatch = oem_relock,
	},
	{
		.command = "format",
		.dispatch = oem_format
	},
	{
		.command = "coda",
		.dispatch = oem_coda
	},
	{
		.command = "setserial",
		.dispatch = oem_setserial
	},
	{
		.command = "setvar",
		.dispatch = oem_setvar
	},
	{
		.command = "ubi",
		.dispatch = oem_ubi,
	},
	{
		.command = "otplock",
		.dispatch = oem_otp_nand_lock,
	},
	{
		.command = "factorymode",
		.dispatch = oem_setfactory,
	}
};

static void handle_oem_command(char *cmd_string, char *response)
{
	int allowed_commands;
	unlock_state_t unlock_state;

	is_unlocked(&unlock_state);
	if (unlock_state < UNLOCK_STATE_RESTRICTED_UNLOCK)
		allowed_commands = OEM_WHITELIST;
	else
		allowed_commands = ARRAY_SIZE(oem_commands);

	if (handle_commands_internal(oem_commands, allowed_commands,
	                             cmd_string, response) < 0) {
		printf("ERROR: oem command %s not recognized.\n", cmd_string);
		fastboot_fail("unrecognized command", response);
	}
}
#endif

/**
 * fastboot_handle_command - Handle fastboot command
 *
 * @cmd_string: Pointer to command string
 * @response: Pointer to fastboot response buffer
 *
 * Return: Executed command, or -1 if not recognized
 */
int fastboot_handle_command(char *cmd_string, char *response)
{
	int ret, allowed_commands;
	unlock_state_t unlock_state;

	is_unlocked(&unlock_state);
	if (unlock_state < UNLOCK_STATE_RESTRICTED_UNLOCK)
		allowed_commands = FB_WHITELIST;
	else
		allowed_commands = ARRAY_SIZE(commands);

	ret = handle_commands_internal(commands,
	                                   allowed_commands,
	                                   cmd_string,
	                                   response);

	if (ret < 0) {
		printf("ERROR: command %s not recognized.\n", cmd_string);
		fastboot_fail("unrecognized command", response);
	}
	return ret;
}

static int handle_commands_internal(const struct cmd_entry *cmds,
                                    int num_commands,
                                    char *cmd_string,
                                    char *response)
{
	int i;
	char *cmd_param;

	for (i = 0; i < num_commands; i++) {
		if (!strncmp(cmds[i].command, cmd_string, strlen(cmds[i].command))) {
			if (cmds[i].dispatch) {
				cmd_param = cmd_string + strlen(cmds[i].command);
				/* handle space or : separators */
				if (cmd_param[0] == ':' || cmd_param[0] == ' ')
					cmd_param++;
				cmds[i].dispatch(cmd_param, response);
				return i;
			} else {
				break;
			}
		}
	}
	return -1;
}


/**
 * okay() - Send bare OKAY response
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 *
 * Send a bare OKAY fastboot response. This is used where the command is
 * valid, but all the work is done after the response has been sent (e.g.
 * boot, reboot etc.)
 */
static void okay(char *cmd_parameter, char *response)
{
	fastboot_okay(NULL, response);
}

/**
 * getvar() - Read a config/version variable
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */
static void getvar(char *cmd_parameter, char *response)
{
	fastboot_getvar(cmd_parameter, response);
}

/**
 * fastboot_download() - Start a download transfer from the client
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */
static void download(char *cmd_parameter, char *response)
{
	char *tmp;

	if (!cmd_parameter) {
		fastboot_fail("Expected command parameter", response);
		return;
	}
	fastboot_bytes_received = 0;
	fastboot_bytes_expected = simple_strtoul(cmd_parameter, &tmp, 16);
	if (fastboot_bytes_expected == 0) {
		fastboot_fail("Expected nonzero image size", response);
		return;
	}

	/*
	 * Nothing to download yet. Response is of the form:
	 * [DATA|FAIL]$cmd_parameter
	 *
	 * where cmd_parameter is an 8 digit hexadecimal number
	 */

	if (fastboot_bytes_expected > fastboot_buf_size) {
		printf("%s failed, expected %d buf %d\n", __FUNCTION__,
			fastboot_bytes_expected, fastboot_buf_size);
		fastboot_fail(cmd_parameter, response);
	} else {
		printf("Starting download of %d bytes\n",
		       fastboot_bytes_expected);
		fastboot_response("DATA", response, "%s", cmd_parameter);
	}
}

/**
 * fastboot_data_remaining() - return bytes remaining in current transfer
 *
 * Return: Number of bytes left in the current download
 */
u32 fastboot_data_remaining(void)
{
	return fastboot_bytes_expected - fastboot_bytes_received;
}

/**
 * fastboot_data_download() - Copy image data to fastboot_buf_addr.
 *
 * @fastboot_data: Pointer to received fastboot data
 * @fastboot_data_len: Length of received fastboot data
 * @response: Pointer to fastboot response buffer
 *
 * Copies image data from fastboot_data to fastboot_buf_addr. Writes to
 * response. fastboot_bytes_received is updated to indicate the number
 * of bytes that have been transferred.
 *
 * On completion sets image_size and ${filesize} to the total size of the
 * downloaded image.
 */
void fastboot_data_download(const void *fastboot_data,
			    unsigned int fastboot_data_len,
			    char *response)
{
#define BYTES_PER_DOT	0x20000
	u32 pre_dot_num, now_dot_num;

	if (fastboot_data_len == 0 ||
	    (fastboot_bytes_received + fastboot_data_len) >
	    fastboot_bytes_expected) {
		fastboot_fail("Received invalid data length",
			      response);
		return;
	}
	/* Download data to fastboot_buf_addr */
	memcpy(fastboot_buf_addr + fastboot_bytes_received,
	       fastboot_data, fastboot_data_len);

	pre_dot_num = fastboot_bytes_received / BYTES_PER_DOT;
	fastboot_bytes_received += fastboot_data_len;
	now_dot_num = fastboot_bytes_received / BYTES_PER_DOT;

	if (pre_dot_num != now_dot_num) {
		putc('.');
		if (!(now_dot_num % 74))
			putc('\n');
	}
	*response = '\0';
}

/**
 * fastboot_data_complete() - Mark current transfer complete
 *
 * @response: Pointer to fastboot response buffer
 *
 * Set image_size and ${filesize} to the total size of the downloaded image.
 */
void fastboot_data_complete(char *response)
{
	char filesz[16];
	/* Download complete. Respond with "OKAY" */
	fastboot_okay(NULL, response);
	printf("\ndownloading of %d bytes finished\n", fastboot_bytes_received);
	image_size = fastboot_bytes_received;
	sprintf(&filesz[0], "%d", image_size);
	setenv("filesize", filesz);
	fastboot_bytes_expected = 0;
	fastboot_bytes_received = 0;
}

#ifdef CONFIG_FASTBOOT_FLASH

#ifdef CONFIG_FASTBOOT_FLASH_NAND
/*
 * Set the UBI partition to search for the named volume, only "UBI" exists
 * at present but we may need to extend this API.
 */

static int mount_ubi_partition(char *cmd_parameter)
{
	/* TODO: What's the VID header all about? */
	if (ubi_part(cmd_parameter, (char *)NULL)) {
		printf("Cannot find UBI partition %s\n", cmd_parameter);
		return -1;
	}

	return 0;
}

/*
 * Extract the volume name from the part:vol string.
 */

#define UBIPREFIX "ubi#"
#define CISNAME "cis"
#define UNLOCKSIG "unlock"
#define FORCEPREFIX "force"

static char *get_ubi_vol_name(char *cmd_parameter)
{
	char *rv;

	rv = cmd_parameter;
	rv += strlen(UBIPREFIX);

	return rv;
}

/*
 * determine if the partition name is CIS, meaning CIS partition table space
 */

static int is_CIS_partition(char *cmd_parameter)
{
	if (!strncmp(cmd_parameter,CISNAME, strlen(CISNAME))){
		printf("Detected CIS partition ID = %s\n", cmd_parameter);
		return 1;
	}
	return 0;
}

/*
 * determine if the partition name is actually an UBI volume that exists
 */

static int is_ubi_volume(char *cmd_parameter)
{
	if (strncmp(cmd_parameter,UBIPREFIX, strlen(UBIPREFIX))){
		return 0;
	}

	printf("Found %s\n", UBIPREFIX);

	if (mount_ubi_partition("UBI")) {
		printf("Can't find UBI partition\n");
		return 0;
	}

	return 1;
}

static int is_unlock_signature(char *cmd_parameter)
{
	if (!strncmp(cmd_parameter,UNLOCKSIG, strlen(UNLOCKSIG))){
		return 1;
	}
	return 0;
}

static int is_force_flash(char **cmd_parameter)
{
	if (strncmp(*cmd_parameter, FORCEPREFIX, sizeof(FORCEPREFIX) - 1) == 0) {
		if ((*cmd_parameter)[sizeof(FORCEPREFIX) - 1] == ':') {
			*cmd_parameter += sizeof(FORCEPREFIX);
			return 1;
		}
	}
	return 0;
}
/*
 * Uses the uBoot MStar writecis command to flash the CIS space.
 * Expects the first 512 bytes of the fastboot buffer to contain the PNI data
 * And the SPINANDINFO data to start at buffer + 0x200 bytes address
 */
static int write_cis(void)
{
	char * const argv[] = {"not_used","0x21000200","0x21000000"};

	printf("Calling writecis 0x%p 0x%p\n",fastboot_buf_addr + 0x200,
					fastboot_buf_addr);

	writeSpinandCIS(NULL, 1, 3, argv);
	return 0;
}

static int write_ubi_volume(char *cmd_parameter)
{
	char *vol_name = get_ubi_vol_name(cmd_parameter);
	int ret;

	/* write UBI volume, should have selected UBI partition already */
	ret = ubi_volume_write(vol_name, fastboot_buf_addr, image_size);

	printf("UBI volume %s write = %d\n", vol_name, ret);

	return ret;
}
#endif

/**
 * flash() - write the downloaded image to the indicated partition.
 *
 * @cmd_parameter: Pointer to partition name
 * @response: Pointer to fastboot response buffer
 *
 * Writes the previously downloaded image to the partition indicated by
 * cmd_parameter. Writes to response.
 */
static void flash(char *cmd_parameter, char *response)
{
	int rv;

#ifdef CONFIG_FASTBOOT_FLASH_NAND
	unlock_state_t unlock_state;

	if (is_unlock_signature(cmd_parameter)) {
		printf("Flashing unlock signature\n");
		rv = write_unlock_signature(fastboot_buf_addr, image_size);
		if (rv) {
			fastboot_fail("Writing unlock signature failed", response);
			return;
		} else {
			printf("Writing unlock signature succeeded\n");
			fastboot_okay(NULL, response);
			return;
		}
	}

	is_unlocked(&unlock_state);
	if (unlock_state < UNLOCK_STATE_RESTRICTED_UNLOCK) {
		printf("Flashing not allowed on locked device\n");
		fastboot_fail("Flashing not allowed!", response);
		return;
	}

	if (is_CIS_partition(cmd_parameter)) {
		printf("Flashing CIS partition table\n");
		if ((rv = write_cis())){
			printf("Write CIS failed %d\n", rv);
			fastboot_fail("Write CIS failed", response);
			return;
		} else {
			printf("Write CIS OK\n");
			fastboot_okay(NULL, response);
			return;
		}
	} else if (is_ubi_volume(cmd_parameter)) {
		printf("Flashing UBI NAND partition\n");
		if ((rv = write_ubi_volume(cmd_parameter))) {
			printf("Write UBI volume failed %d\n", rv);
			fastboot_fail("Write UBI failed", response);
			return;
		} else {
			printf("Write UBI volume succeeded\n");
			fastboot_okay(NULL, response);
			return;
		}
	} else if (is_otp_region(cmd_parameter)) {
		printf("Flashing NAND OTP region\n");
		if ((rv = write_nand_otp(cmd_parameter,fastboot_buf_addr,image_size))) {
			printf("Write NAND OTP failed %d\n", rv);
			fastboot_fail("Write OTP failed", response);
			return;
		} else {
			printf("Write NAND OTP succeeded\n");
			fastboot_okay(NULL, response);
			return;
		}
	} else {
		if (is_force_flash(&cmd_parameter)) {
			fastboot_nand_flash_write(cmd_parameter,
			                          fastboot_buf_addr,
			                          image_size, response);
		} else {
			fastboot_nand_flash_write_if_modified(cmd_parameter,
			                                      fastboot_buf_addr,
			                                      image_size, response);
		}
	}
#endif
#ifdef CONFIG_FASTBOOT_FLASH_MMC
	fastboot_mmc_flash_write(cmd_parameter, fastboot_buf_addr, image_size,
				 response);
#endif
}

/**
 * erase() - erase the indicated partition.
 *
 * @cmd_parameter: Pointer to partition name
 * @response: Pointer to fastboot response buffer
 *
 * Erases the partition indicated by cmd_parameter (clear to 0x00s). Writes
 * to response.
 */
static void erase(char *cmd_parameter, char *response)
{
#ifdef CONFIG_FASTBOOT_FLASH_MMC
	fastboot_mmc_erase(cmd_parameter, response);
#endif
#ifdef CONFIG_FASTBOOT_FLASH_NAND
	fastboot_nand_erase(cmd_parameter, response);
#endif
}
#endif

/*
 * reboot_bootloader() - Sets reboot bootloader flag.
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */
static void reboot_bootloader(char *cmd_parameter, char *response)
{
	if (fastboot_set_reboot_flag())
		fastboot_fail("Cannot set reboot flag", response);
	else
		fastboot_okay(NULL, response);
}

#ifdef CONFIG_FASTBOOT_CMD_OEM
/**
 * oem_coda() - Reset the MT7686 into CODA mode for flashing
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 * NB: MStar GPIO API is homegrown not the standard uBoot.
 */

/*
 * SAV GPIO55 connects to MTK GPIO12, selects MTK UART download if low at boot.
 * SAV GPIO79 connects to MTK !CE, low for MTK to start, high holds in reset.
 * Need to leave 20mS delay for MTK to actually reset.
 */

#define GPIO_14 14
#define GPIO_15 15
#define GPIO_55 55
#define GPIO_79 79
#define GPIO_66 66
#define PWR_CYCLE_DELAY_US 20000

static void oem_coda(char *cmd_parameter, char *response)
{
	/* Set GPIO66 high to keep SAV526 powered */
	printf("Enabling blue power\n");
	MDrv_GPIO_Pad_Set(GPIO_66);
	mdrv_gpio_set_high(GPIO_66);

	/* Set GPIO55 low to enter CODA mode */
	MDrv_GPIO_Pad_Set(GPIO_55);
	mdrv_gpio_set_low(GPIO_55);

	printf("Reset red\n");
	MDrv_GPIO_Pad_Set(GPIO_14);
	mdrv_gpio_set_input(GPIO_14);
	MDrv_GPIO_Pad_Set(GPIO_15);
	mdrv_gpio_set_input(GPIO_15);

	/* Toggle the GPIO79 to high then low to reset the MT76XX */
	MDrv_GPIO_Pad_Set(GPIO_79);
	mdrv_gpio_set_high(GPIO_79);
	udelay(PWR_CYCLE_DELAY_US);
	mdrv_gpio_set_low(GPIO_79);

	/* Apparently cannot fail!! */
	fastboot_okay(NULL, response);
}

/**
 * Set the UBI partition (mount MTD partition as UBI)
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */

static void set_ubi_partition(char *cmd_parameter, char *response)
{
	char cmdbuf[32], *arg1;

	strtok(cmd_parameter, " ");
	arg1 = strtok(NULL," ");

	printf("%s input %s arg %s\n", __FUNCTION__, cmd_parameter, arg1);

	sprintf(cmdbuf, "ubi part %s", arg1);
	printf(cmdbuf, "ubi part %s", arg1);
	if (run_command(cmdbuf, 0))
		fastboot_fail("", response);
	else
		fastboot_okay(NULL, response);
}

/**
 * create_ubi_vol() - Create an empty UBI volume
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */

static void create_ubi_vol(char *cmd_parameter, char *response)
{
	char cmdbuf[32], *name, *size;

	printf("%s arg %s\n", __FUNCTION__, cmd_parameter);

	strtok(cmd_parameter, " ");
	name = strtok(NULL, " ");
	size = strtok(NULL, " ");

	printf("name %s size %s\n", name, size);

	sprintf(cmdbuf, "ubi create %s %s", name, size);

	printf("command %s\n", cmdbuf);
	if (run_command(cmdbuf, 0))
		fastboot_fail("", response);
	else
		fastboot_okay(NULL, response);
}

static void remove_ubi_vol(char *cmd_parameter, char *response)
{
	char cmdbuf[32], *volname;

	printf("%s arg %s\n", __FUNCTION__, cmd_parameter);

	strtok(cmd_parameter, " ");
	volname = strtok(NULL, " ");

	sprintf(cmdbuf, "ubi remove %s", volname);
	if (run_command(cmdbuf, 0))
		fastboot_fail("", response);
	else
		fastboot_okay(NULL, response);
}

static void run(char *cmd_parameter, char *response)
{
	char cmdbuf[32];

	sprintf(cmdbuf, "%s", cmd_parameter);
	printf("Running cmdline: %s\n",cmdbuf);
	if (run_command(cmdbuf, 0))
		fastboot_fail("", response);
	else
		fastboot_okay(NULL, response);
}

/*
 * To allow working from empty flash (and making one)
 * 1) erase UBI partition "oem ubi clear partition_name"
 * 2) ubi part <partition_name> "oem ubi mount partition_name"
 * 	NB! Something needs to allow ubi part UBI to really mount UBI partition.
 *	NB! New UBI command, like "ubi decache"
 * 3) create volume on partition "oem ubi cvol volume_name size"
 */

static void oem_ubi(char *cmd_parameter, char *response)
{
	/* Parse UBI commands from cmd_parameter string */
	if (!strncmp(cmd_parameter,"setpart",strlen("setpart")))
		set_ubi_partition(cmd_parameter, response);
	else if (!strncmp(cmd_parameter,"mkvol",strlen("cvol")))
		create_ubi_vol(cmd_parameter, response);
	else if (!strncmp(cmd_parameter,"rmvol",strlen("evol")))
		remove_ubi_vol(cmd_parameter, response);
	else if (!strncmp(cmd_parameter,"runcmd",strlen("runcmd")))
		run(cmd_parameter, response);
	else
		fastboot_fail("Unrecognised UBI oem command", response);
}

static void oem_setserial(char *cmd_parameter, char *response)
{
	if (setenv("serial#", cmd_parameter)) {
		printf("Failed to set serial# to %s\n", cmd_parameter);
		fastboot_fail("Failed to set serial#", response);
	} else {
		saveenv();
		fastboot_okay(NULL, response);
	}
}

static void oem_setvar(char *cmd_parameter, char *response)
{
	int i=0;
	int paramlen = strlen(cmd_parameter);
	int varlen;
	char* varval;

	for(i = 0; i < paramlen; i++) {
		if ((cmd_parameter[i] == ':') || (cmd_parameter[i] == ' ')) {
			cmd_parameter[i] = '\0';
			/* Allow variables with ' ' and ':'' in them' */
			break;
		}
	}

	varlen = strlen(cmd_parameter);
	if (varlen == paramlen) {
		fastboot_fail("Expected variable name and value", response);
		return;
	}

	varval = cmd_parameter + varlen + 1;

	if (setenv(cmd_parameter, varval)) {
		printf("Failed to set %s to %s\n", cmd_parameter, varval);
		fastboot_fail("Failed to set environment variable", response);
	} else {
		saveenv();
		fastboot_okay(NULL, response);
	}
}

/**
 * oem_format() - Execute the OEM format command
 *
 * @cmd_parameter: Pointer to command parameter
 * @response: Pointer to fastboot response buffer
 */
static void oem_format(char *cmd_parameter, char *response)
{
	char cmdbuf[32]; //PK:TODO: What is running this command? uBoot?

	if (!getenv("partitions")) {
		fastboot_fail("partitions not set", response);
	} else {
		/* sprintf(cmdbuf, "gpt write mmc %x $partitions",
			CONFIG_FASTBOOT_FLASH_MMC_DEV); */
		if (run_command(cmdbuf, 0))
			fastboot_fail("", response);
		else
			fastboot_okay(NULL, response);
	}
}

extern int relock_device(void);
#ifdef CONFIG_FASTBOOT_FLASH_NAND
static void oem_otp_nand_lock(char *cmd_parameter, char *response)
{
	int ret = fb_otp_nand_lock();
	if (ret == 0) {
		fastboot_okay(NULL, response);
		/* Ensure the device is set as locked */
		relock_device();
	} else {
		fastboot_fail("Failed to lock NAND OTP", response);
	}
}
#endif

static void oem_setfactory(char *cmd_parameter, char *response)
{
	int paramlen = strlen(cmd_parameter);
	char *tmp;

	uint32_t port = 0;
	if (paramlen > 0) {
		port = simple_strtoul(cmd_parameter, &tmp, 10);
	}
	if (port) {
		if (enable_factory_test_mode(port) < 0)
			fastboot_fail("Failed to enable factory test mode", response);
		else {
			printf("Set Factory Test Mode with port %d\n", port);
			fastboot_okay(NULL, response);
		}
	} else {
		disable_factory_test_mode();
		printf("Disabled Factory Test Mode\n");
		fastboot_okay(NULL, response);
	}
}
#endif

static void oem_relock(char *cmd_parameter, char *response)
{
	int ret = relock_device();
	if (ret == 0) {
		fastboot_okay(NULL, response);
	} else {
		fastboot_fail("Failed to relock device", response);
	}
}
