/*
* Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
*
* SPDX-License-Identifier:	GPL-2.0+
*/


#include <common.h>
#include <fastboot.h>
#include <fb_otp.h>
#include <fastboot-internal.h>
#include <ring_platform_features.h>
#include <nand.h>
#include <stdlib.h>
#include <spinand.h>
#include <drvSPINAND.h>

/*
 * determine if the partition name is OTP
 */

int is_otp_region(char *cmd_parameter) {
	if (!strncmp(cmd_parameter, OTPNAME, strlen(OTPNAME))) {
		return 1;
	}

	return 0;
}

int write_nand_otp(char *cmd_parameter, void *fastboot_buf_addr, u32 image_size) {
	u32 ret = -1;
	loff_t off = 0;
	int dev = 0;
	nand_info_t *nand = &nand_info[dev];

	if (!is_otp_programmed()) {
		if (image_size > (MAX_PAGE_SZ * MAX_OTP_PAGES)) {
			printf("Image size (%d) larger than max size (%d)",
			    image_size, (MAX_PAGE_SZ * MAX_OTP_PAGES));
			return ret;
		}

		off += MAX_PAGE_SZ * MDrv_SPINAND_GetOTPPageOffset();

		ret = MDrv_SPINAND_EnableOtp(true);
		if (ret) {
			printf("Error switching to OTP mode: %d\n",ret);
			return ret;
		}

		ret = nand_write_skip_bad(nand, off, &image_size,
					NULL, image_size,
					(u_char *)fastboot_buf_addr, 0);

		if (ret)
			printf("RW fail %d\n", ret);

		ret = MDrv_SPINAND_EnableOtp(false);
		if (ret)
			printf("Error switch from OTP mode: %d\n", ret);
	} else {
		uint8_t *written_data = malloc(MAX_PAGE_SZ);
		size_t i;

		ret = 0;
		printf("OTP aleady programmed. Comparing...\n");
		for (i = 0; i < (image_size + (MAX_PAGE_SZ-1)) / MAX_PAGE_SZ; i++) {
			int cmpret = 0;
			u32 compare_size = MAX_PAGE_SZ;
			read_otp_page(i, written_data, MAX_PAGE_SZ, 0);
			if (i == 0) {
				/* Only compare the data in the otp_meta
				   structure.  The rest of the first page is
				   random data, so can only be compared if we
				   have retained the original file */
				compare_size = sizeof(struct otp_meta);
			}
			cmpret = memcmp(written_data,
				     fastboot_buf_addr + (i * MAX_PAGE_SZ),
				     compare_size);
			printf("Page %d %s\n", i, cmpret ? "MISMATCHED!" :
			                                   "Compared OK");
			if (cmpret)
				ret = -1;
		}

		free(written_data);
	}

	return ret;
}

int fb_otp_nand_lock(void)
{
	int ret = 0;
	if (!MDrv_SPINAND_IsOtpLocked()) {
		ret = MDrv_SPINAND_LockOtp();
		if(ret == ERR_SPINAND_SUCCESS)
			ret = 0;
		else
			ret = -1;
	}
	return ret;
}

bool is_otp_programmed() {
	bool err = false;
	static const int load_size = MAX_PAGE_SZ; /* 1 NAND page */
	u8 *tmpkey = malloc(load_size);
	if (!tmpkey) {
		printf("Failed to allocate space to load metadata\n");
		goto out;
	}
	if (read_otp_page(META_DATA_PAGE, tmpkey, load_size, 0)) {
		printf("Failed to read META page from NAND OTP\n");
		goto out_free;
	}
	if(!strncmp((char *)tmpkey,COCOA_MAGIC,strlen(COCOA_MAGIC))) {
		printf("OTP programmed!\n");
		err = true;
	} else
		printf("OTP Not programmed!\n");
out_free:
	free(tmpkey);
out:
	return err;
}

int try_otp_read_dsn(char *dsn) {
	return try_otp_read_all(dsn, NULL);
}

int try_otp_read_all(char *dsn, char *boardid) {
	int err = -1;
	static const int load_size = MAX_PAGE_SZ; /* 1 NAND page */
	u8 *tmpkey = malloc(load_size);
	if (!tmpkey) {
		printf("Failed to allocate space to load metadata\n");
		goto out;
	}
	err = read_otp_page(META_DATA_PAGE, tmpkey, load_size, 0);
	if (err) {
		printf("Failed to read META page from NAND OTP\n");
		goto out_free;
	}
	char *src = (char *)tmpkey;
	if(!strncmp(src,COCOA_MAGIC,strlen(COCOA_MAGIC))) {
		src += strlen(COCOA_MAGIC)+COCOA_VER_LEN;
		if (boardid) {
			strncpy(boardid, src, COCOA_BOARD_ID_LEN);
			boardid[COCOA_BOARD_ID_LEN]='\0';
		}
		src += COCOA_BOARD_ID_LEN;
		if (dsn) {
			strncpy(dsn, src, COCOA_DSN_LEN);
			dsn[COCOA_DSN_LEN]='\0';
		}
		err = 0;
	} else {
		printf("Can't metadata read from OTP\n");
		err = -1;
	}
out_free:
	free(tmpkey);
out:
	return err;
}

int set_serialno(void) {
	int err;
	char dsn[COCOA_DSN_LEN+1];
	err = try_otp_read_dsn(dsn);
	if (err == 0) {
		printf ("DSN from OTP = %s\n",dsn);
		const char *tmp = getenv("serial#");
		if (!tmp) {
			setenv("serial#",dsn);
			saveenv();
		} else {
			if (strncmp(tmp,dsn,COCOA_DSN_LEN)) {
				printf("The DSN from OTP & env mismatch!\n");
				setenv("serial#",NULL);
				setenv("serial#",dsn);
				saveenv();
			} else
				printf("DSN in OTP and ENV are same\n");
		}
	} else
		printf("Can't read DSN, OTP not programmed\n");

	return err;

}
